// ESFrameModelSwarm.java

import swarm.Globals;
import swarm.Selector;
import swarm.defobj.Zone;
import swarm.defobj.SymbolImpl;

import swarm.activity.Activity;
import swarm.activity.ActionGroup;
import swarm.activity.ActionGroupImpl;
import swarm.activity.Schedule;
import swarm.activity.ScheduleImpl;

import swarm.objectbase.Swarm;
import swarm.objectbase.SwarmImpl;
import swarm.objectbase.VarProbe;
import swarm.objectbase.MessageProbe;
import swarm.objectbase.EmptyProbeMapImpl;

import swarm.collections.ListImpl;
import swarm.collections.ListIndex;
import swarm.collections.ArrayImpl;

import java.io.*;
import java.util.*;

/**
 * The Model contains the Units and all the related tools, like the
 * warehouses which are necessary to deal with the inventories or the
 * newses
 *
 * @author Pietro Terna, with contributions of Marco Remondino, Dario Landini, 
 * Nicola Ormezzano and Paolo Pelligra
 */
public class ESFrameModelSwarm extends SwarmImpl
{
    // simulation parameters

    /** the number of ticks of the clock in a day; at each tick a unit
     * of each step in recipes is done 
     */ 
    public int ticksInATimeUnit;
    /** the total number of operating Units we are using*/
    public int totalUnitNumber;
    /** the total number of end units we are using; and endUnit represents
     *  an actual or virtual warehouse were we can "find" components 
     *  internallly produced using sub-recipes or external procurements; 
     *  each endUnit has a code which is the same used to conclude (as 
     *  an 'end') the recipe describing this internal or external 
     *  production */
    public int totalEndUnitNumber;
    /** the total number of layers;      
     *  a layer is a period of time or a set of qualitative conditions that 
     *  introduce differentiations into the orders; e.g two collections in 
     *  fashion production, with the same technical description (recipe) and 
     *  different qualitative results */
    public int totalLayerNumber;
    /** the max number of steps to be done to complete an order*/
    public int maxStepNumber;
    /** the max length of a step in a recipe (in units of time or ticks) */
    public int maxStepLength;
    /** the max number of ticks that a product can spent in a unit */
    public int maxTickInAUnit;
    /** the max inventory quantity in a warehouse */
    public int maxInWarehouses;
    /** the min inventory quantity in a warehouse */
    public int minInWarehouses;
    /** choosing if we use or not warehouses (and inventories) */
    public boolean useWarehouses;
    /** choosing if we use or not newses  */
    public boolean useNewses;
    /** how deep information are propagated (how many steps we examine after
     * the current production phase)*/
    public int infDeepness;
    /** the estimate of the revenue of the enterprise per each recipe step 
     * done in a semi manufactured order or in a final order */
    public float revenuePerEachRecipeStep;
    /** the coefficient to be applied to costs to estimate the revenue of the enterprise 
     *  per unity of cost (to be used in an alternatice way with the previous solution) */
    public float revenuePerCostUnit;
    /** the yearly rate used to evaluate the financial costs of inventories */
     public  float inventoryFinancialRate;

    /** the number of newses necessary to decide the production of 
     *  inventories */
    public int nOfNewsesToProduce;
    /** the number of newses to be cleared after the decision of producing
     *  to increase inventories (a simple choice is to have it equal to 
     *  nOfNewsesToProduce) */
    public int nOfNewsesToBeCleared;
    /** the number of orders for which newses are propagated by a unit */
    public int nOfOrdersInNewses;
    /** the inventory evaluation criterion */
    public  int inventoryEvaluationCriterion;
    /** the total production time used by done orders */
    public float totalProductionTime;
    /** the total length of the recipes of the done orders */
    public float totalRecipeLength;

    /** the order generator, to be replaced by the order distiller or 
     *  in the future by more sophisticated structures (swarms of 
     *  enterprises etc.) */
    public OrderGenerator orderGenerator;

    /** use the order distiller instead of the order generator */
    public boolean useOrderDistiller;

    /** the order distiller, using archives of recipes and of order sequences
     *  to generate the simulated events */
    public OrderDistiller orderDistiller;

    /** ActionGroup for holding an ordered sequence of action */
    public ActionGroup modelActions1, modelActions2, modelActions2b, 
	modelActions2generator, modelActions2distiller, modelActions3;

    /** the Schedule operating in the Model*/
    public Schedule modelSchedule;

    /** the inventoryRuleMaster managing inventories */
    public InventoryRuleMaster inventoryRuleMaster;

    /** the informationRuleMaster managing newses */
    public InformationRuleMaster informationRuleMaster;
    /** the input class for unit parameters*/
    public UnitParameters unitParameters;

    /** the list of the operating units */
    public ListImpl unitList;
    /** its iterator */
    public ListIndex unitListIndex;
    /** the list of the warehouses */
    public ListImpl warehouseList;
    /** the whole list of the orders (not yet finished) */
    public ListImpl orderList;
    /** its iterator */
    public ListIndex orderListIndex;

    /** the list of the end units */
    public ListImpl endUnitList;

    /** the list of procurementAssembler units linked to normal unit; 
     *  proc. ass. units have in charge the task of making the procurement */
    public ListImpl procurementAssemblerList;

    /** the list of computationalAssembler units linked to normal unit; 
     *  comp. ass. units have in charge the task of making the computations */
    public ListImpl computationalAssemblerList;

    /** the list of standAloneBatchAssembler units linked to normal unit */
    public ListImpl standAloneBatchAssemblerList;

    /** the list of sequentialBatchAssembler units linked to normal unit */
    public ListImpl sequentialBatchAssemblerList;

    /** the log of the concluded orders */
    public MyWriter concludedOrderLog=null;

    /** the criterion used dealing with 'or' sequences */
    public int orCriterion;

    /** the criterion used to choose the subsequent unit when more the one unit
     *  is able to perform the required task */
    public int unitCriterion;

    /** the tool to assign order to units */
    AssigningTool assigningTool;

    /** unused variable, kept for backward compatibility with jesframe.scm files
     *  older then the jes versione 0.9.7.61 */
    public int distillerMultiplicity;

    /** do not account for costs for the whole first bunck of tick (the number of tick
     *  defined as ticksInATimeUnit) */
    public boolean noAccountingInFirstTimeUnit;

    /** the array of memory matrixes to be used by computational objects */
    public ArrayImpl matrixArray;

    /** a memory matrix */
    public MemoryMatrix aMatrix;

    /** printMatrixes */
    public boolean printMatrixes;

    /** number of memory matrixes */
    public int totalMemoryMatrixNumber;

    /** the number of the memory matrix to be uses in 'or' 
     *  processes (if any) */
    public int orMemoryMatrix;

    /** using the LIFO reassignment of order if the next step
     *  is the same as the previous one */
    public boolean sameStepLifoAssignment;

    /** assigning an order to the unit it come from if 
     *  if the step to be done is the same as the previous one */
    public boolean assignEqualStepsToSameUnit;

    /** in sequential batch two orders can stay in the same batch
     *  if several criteria match; one of them is that of having the
     *  same state vector, which contains the number of the units
     *  that made the various steps; if we consider more simply the state
     *  as 0 or non 0, we disregard the unit number (also the 'or', case
     *  with -1 values in not used branches, receives the same treatment) */
    public boolean compareDisregardingUnits;

    /** the address of the probe, if we need to access it; remember casting
     *  (ESFrameModelSwarm$1$ESFrameModelProbeMap)
     */
    Object modelProbe;

    // the constructor of the model
    public ESFrameModelSwarm (Zone aZone) {
	super (aZone);

	// Now fill in various simulation parameters with default values.
	useOrderDistiller=false;
	ticksInATimeUnit = 1;
	totalUnitNumber = 10;
	totalEndUnitNumber = 3;
	totalLayerNumber=1;
	totalMemoryMatrixNumber=4;
	maxStepNumber=30;
	maxStepLength=1;
	sameStepLifoAssignment=false;
	assignEqualStepsToSameUnit=false;
	compareDisregardingUnits=false;
	maxTickInAUnit=0;
	useWarehouses=true;
	maxInWarehouses=3;
	minInWarehouses=3;
	useNewses=true;
	infDeepness=3;
	inventoryFinancialRate= (float) 0.1;
	inventoryEvaluationCriterion=1;
	revenuePerEachRecipeStep=2;
	revenuePerCostUnit=0;
	nOfNewsesToProduce=1;
	nOfNewsesToBeCleared=100000;
	nOfOrdersInNewses=1;
	orCriterion=0;
	orMemoryMatrix=2;
	unitCriterion=0;
	distillerMultiplicity=1; // unused variable, kept for backward compatibility
	                         // with old jesframe.scm files
	noAccountingInFirstTimeUnit=false;

	// the log of the orders
	try{
	    concludedOrderLog = new MyWriter("log/concludedOrderLog.txt");
	}
	catch (IOException e){
	    System.out.println(e);
	    MyExit.exit(1);
	}
	String text = 
	"Each line contains: final time unit; tick in the final time unit;\n"+
	"                    recipe name; order layer; order number;\n"+
	"                    starting time unit; tick in the s. time unit\n"+
	"                    (number of steps); the recipe steps\n"+
	"                    { the units that have been doing the various"+
                               " steps of the\n"+
        "                    order (-1=step not executed, 'or' sequence) }\n"+
	"                    [total final cost of the order]\n"+
        "                    [[number of direct and indirect steps done]]\n"+
        "                    [order final evaluation]"+
                             " multiplicity\n";
	try{
	    concludedOrderLog.write(text);}
	catch(IOException e){
	    System.out.println(e);
	    MyExit.exit(1);
			}

	// Build a customized probe map using a `local' subclass
	// (a special kind of Java `inner class') of the
	// EmptyProbeMapImpl class.
	// This building operation produces a separated class file, with
	// name  ESFrameModelSwarm$1$ESFrameModelProbeMap.class
	class ESFrameModelProbeMap extends EmptyProbeMapImpl {
	    public VarProbe probeVariable (String name) {
		return
		    Globals.env.probeLibrary.getProbeForVariable$inClass
		    (name, ESFrameModelSwarm.this.getClass ());
	    }
	    public MessageProbe probeMessage (String name) {
		return
		    Globals.env.probeLibrary.getProbeForMessage$inClass
		    (name, ESFrameModelSwarm.this.getClass ());
	    }
	    public void addVar (String name) {
		addProbe (probeVariable (name));
	    }
	    public void addMessage (String name) {
		addProbe (probeMessage (name));
	    }
	    public ESFrameModelProbeMap (Zone _aZone, Class aClass) {
		super (_aZone, aClass);
		addVar ("ticksInATimeUnit");
		addVar ("totalUnitNumber");
	    	addVar ("totalEndUnitNumber");
		addVar ("totalLayerNumber");
		addVar ("totalMemoryMatrixNumber");
		addVar ("sameStepLifoAssignment");
		addVar ("assignEqualStepsToSameUnit");
		addVar ("compareDisregardingUnits");
		addVar ("maxTickInAUnit");
		addVar ("useWarehouses");
		addVar ("useNewses");
		addVar ("maxInWarehouses");
		addVar ("minInWarehouses");
		addVar ("infDeepness");
		addVar ("inventoryFinancialRate");
		addVar ("inventoryEvaluationCriterion");
		addVar ("revenuePerEachRecipeStep");
		addVar ("revenuePerCostUnit");
		addVar ("nOfNewsesToProduce");
		addVar ("nOfNewsesToBeCleared");
		addVar ("nOfOrdersInNewses");
		addVar ("orCriterion");
		addVar ("orMemoryMatrix");
		addVar ("unitCriterion");
		addVar ("noAccountingInFirstTimeUnit");
		addVar ("useOrderDistiller");
		addVar ("maxStepNumber");
		addVar ("maxStepLength");
    }
	}

	// Install our custom probeMap class directly into the
	// probeLibrary
	modelProbe=new ESFrameModelProbeMap (aZone, getClass ());
	Globals.env.probeLibrary.setProbeMap$For
	    ((ESFrameModelProbeMap) modelProbe, getClass ());
    }
    /** Build the model objects. */
    public Object buildObjects ()
    {
	int i;

	// allow our parent class to build anything
	super.buildObjects();

        // create the class to read acounting parameters
        unitParameters = new 
	    UnitParameters(getZone(), totalUnitNumber, totalEndUnitNumber);
        // read fixed and variable costs
        unitParameters.readUnitData();

	// check maxInWarehouses value and if 0 change it to the max
	// integer value
	if (maxInWarehouses==0)
	    {maxInWarehouses=Integer.MAX_VALUE;
	    // no verbose control here
	    System.out.println(
      "Warning, maxInWarehouses has been changed from 0 to Integer.MAX_VALUE");
	    }

	// a warning about the use of the warehouses if layers are more 
	// than one
	if(totalLayerNumber>1 && useWarehouses)
	    System.out.println(
			       "You are using warehouses as stand alone "+
			       "production of inventories.\nWarning: At "+
			       "present it is not possible to check "+
			       "the consistence of the\ninventory layer and "+
			       "that of an order. (This is a temporary "+
			       "limitation).");

	// check the layer number
	if(totalLayerNumber<=0){
	    System.out.println("The number of layer cannot be <=0.");
	    MyExit.exit(1);
	}

	// the list of the production units
	unitList = new ListImpl (getZone());
	// its iterator
	unitListIndex = unitList.listBegin(getZone());
	// the list of the warehouses
	warehouseList = new ListImpl (getZone());
        // the whole list of the orders existing at each time step
        orderList = new ListImpl(getZone());
	// its iterator
	orderListIndex = orderList.listBegin(getZone());
	// the list of the end units
	endUnitList = new ListImpl (getZone());

	// the instance of assigningTool
	assigningTool = new AssigningTool (getZone(), unitCriterion,
					   unitList, endUnitList,
					   assignEqualStepsToSameUnit);

	// setting 'or' criterion and unitList address in 'Or'
	Or.setOrCriterion(orCriterion);
	Or.setUnitList(unitList);

	// the list of the procurement assemblers linked to the units 
	// that will be generated 
	procurementAssemblerList = new ListImpl (getZone());

	// the list of the computational assemblers linked to the units 
	// that will be generated 
	computationalAssemblerList = new ListImpl (getZone());

	// the list of the stand alone batch assemblers linked to the units 
	// that will be generated 
	standAloneBatchAssemblerList = new ListImpl (getZone());

	// the list of the sequential batch assemblers linked to the units 
	// that will be generated 
	sequentialBatchAssemblerList = new ListImpl (getZone());

	// the orderGenerator or the orderDistiller
	if(! useOrderDistiller)
	    orderGenerator = new OrderGenerator
		(getZone(), maxStepNumber, maxStepLength, unitList, 
		 endUnitList, orderList, totalLayerNumber, this, 
		 assigningTool);
	else
	    orderDistiller = new OrderDistiller
		(getZone(), maxStepNumber, maxStepLength, unitList, 
		 endUnitList, orderList, totalLayerNumber, this, 
		 assigningTool);

	// the inventoryRuleMaster managing the inventories
	inventoryRuleMaster = new 
	    InventoryRuleMaster(getZone(),
				maxInWarehouses, minInWarehouses,
				nOfNewsesToProduce, nOfNewsesToBeCleared);

        // the informationRuleMaster managing the news propagation
        if(useWarehouses && useNewses)
        informationRuleMaster = new InformationRuleMaster(getZone(),
							  totalUnitNumber);

	// create the endUnits
	for (i = 1; i <= totalEndUnitNumber; i++)
	    {
		EndUnit anEndUnit;

		anEndUnit  =  new EndUnit (getZone(), this, totalLayerNumber);
		unitParameters.setParametersToEndUnit(i, anEndUnit);

		endUnitList.addLast (anEndUnit);
		}

	// create the units and their warehouses, if any
	for (i = 1; i <= totalUnitNumber; i++)
	    {
		Warehouse aWarehouse=null; // null to avoid errors when unused
		News aNews=null;           //  "   "   "     "      "    "
		Unit aUnit;
		ProcurementAssembler aProcurementAssembler;
		ComputationalAssembler aComputationalAssembler;
		StandAloneBatchAssembler aStandAloneBatchAssembler;
		SequentialBatchAssembler aSequentialBatchAssembler;

		// using warehouses and not newses
		if(useWarehouses && ! useNewses)
		{
		    // using the unit number as id also for its warehouse
		    aWarehouse =  new Warehouse (getZone(), unitParameters.
						 getUnitNumber(i));
		    aProcurementAssembler = new ProcurementAssembler 
			(getZone(), endUnitList);
		    aComputationalAssembler = new ComputationalAssembler 
			(getZone());
		    aStandAloneBatchAssembler = new 
			StandAloneBatchAssembler (getZone());
		    aSequentialBatchAssembler = new 
			SequentialBatchAssembler (getZone(), 
						  compareDisregardingUnits);
		    aUnit  =  new Unit (getZone(), unitList, endUnitList, this,
					inventoryRuleMaster,  aWarehouse,
					orderList, aProcurementAssembler,
					aComputationalAssembler,
					aStandAloneBatchAssembler,
					aSequentialBatchAssembler,
					concludedOrderLog, assigningTool,
					maxTickInAUnit, 
					sameStepLifoAssignment,
					noAccountingInFirstTimeUnit);
		    unitParameters.setParametersTo(i, aUnit);
		    aProcurementAssembler.setUnit(aUnit);
		    aComputationalAssembler.setUnit(aUnit);
		    aStandAloneBatchAssembler.setUnit(aUnit);
		    aSequentialBatchAssembler.setUnit(aUnit);
		}
		// using both warehoueses and newses
		else if(useWarehouses && useNewses)
		    {
			// using the unit number as id also for its warehouse
			aWarehouse =  new Warehouse (getZone(), unitParameters.
						     getUnitNumber(i));
			aProcurementAssembler = new ProcurementAssembler 
			    (getZone(), endUnitList);
			aComputationalAssembler = new ComputationalAssembler 
			    (getZone());
			aStandAloneBatchAssembler = new 
			    StandAloneBatchAssembler (getZone());
			aSequentialBatchAssembler = new 
			    SequentialBatchAssembler (getZone(), 
						  compareDisregardingUnits);
			aUnit  =  new Unit(getZone(), unitList, endUnitList,
					   this,
					   inventoryRuleMaster,  aWarehouse,
					   informationRuleMaster, infDeepness, 
				           nOfOrdersInNewses,
				           orderList, aProcurementAssembler,
					   aComputationalAssembler,
					   aStandAloneBatchAssembler,
					   aSequentialBatchAssembler,
					   concludedOrderLog, assigningTool,
					   maxTickInAUnit,
					   sameStepLifoAssignment,
					   noAccountingInFirstTimeUnit);
			unitParameters.setParametersTo(i, aUnit);
			aProcurementAssembler.setUnit(aUnit);
			aComputationalAssembler.setUnit(aUnit);
			aStandAloneBatchAssembler.setUnit(aUnit);
			aSequentialBatchAssembler.setUnit(aUnit);
		    }
		// no warehouses and no newses
		else 
		    {
			aProcurementAssembler = new ProcurementAssembler 
			    (getZone(), endUnitList);
			aComputationalAssembler = new ComputationalAssembler 
			    (getZone());
			aStandAloneBatchAssembler = new 
			    StandAloneBatchAssembler (getZone());
			aSequentialBatchAssembler = new 
			    SequentialBatchAssembler (getZone(), 
						  compareDisregardingUnits);
			aUnit = new Unit  (getZone(), unitList, endUnitList, 
					   this, orderList, 
					   aProcurementAssembler,
					   aComputationalAssembler,
					   aStandAloneBatchAssembler,
					   aSequentialBatchAssembler,
					   concludedOrderLog, assigningTool,
					   maxTickInAUnit,
					   sameStepLifoAssignment,
					   noAccountingInFirstTimeUnit);
			unitParameters.setParametersTo(i, aUnit);
			aProcurementAssembler.setUnit(aUnit);
			aComputationalAssembler.setUnit(aUnit);
			aStandAloneBatchAssembler.setUnit(aUnit);
			aSequentialBatchAssembler.setUnit(aUnit);
		    }
		// add the unit, the assembler and the warehouse to the lists
		unitList.addLast (aUnit);
		if(useWarehouses) warehouseList.addLast (aWarehouse);
		procurementAssemblerList.addLast(aProcurementAssembler);
		computationalAssemblerList.addLast(aComputationalAssembler);
		standAloneBatchAssemblerList.
		    addLast(aStandAloneBatchAssembler);
		sequentialBatchAssemblerList.
		    addLast(aSequentialBatchAssembler);

	    }

	if(! useWarehouses)if(StartESFrame.verbose)
	    System.out.println("No warehouses generated");


	// creating the dictionary in orderGenerator or in OrderDistiller
	if(! useOrderDistiller)	orderGenerator.setDictionary();
	else                    orderDistiller.setDictionary();

	// creating memory matrixes
	if(totalMemoryMatrixNumber>0){

	matrixArray = new ArrayImpl(getZone(), totalMemoryMatrixNumber);


	// reading memory matrix dimension from unitData/memoryMatrixes.txt
	// and checking the sequence of the data (the file must contains
	// the data in a orderd way; matrixes are identified by a natural 
	// integer number sequence, starting from 0

	try{

	    // read the file
	    BufferedReader fileMemoryMatrixes = null;
	    StringTokenizer t;
	    int matrixNumber, rows, cols;
	    boolean sensitive;
	    fileMemoryMatrixes = new 
		BufferedReader(new FileReader("unitData/memoryMatrixes.txt"));

	    // reading a control line (titles)
	    String line = " ", lineTest =
         "number(from_0_ordered;_negative_if_insensitive_to_layers)_rows_cols";

	    line=fileMemoryMatrixes.readLine();
	    if (! line.equals(lineTest))
		{
		    System.out.println(
		"First line in unitData/unitMemoryMatrixes.txt must contain:");
		    System.out.println(lineTest);
		    MyExit.exit(1);
		}
	    line=" ";

	    // starting with a void line; the actual
	    // reading is done in MyReader, dealing also
	    // with the missing data case
	    t = new StringTokenizer(line, " ");

	    for(i = 0; i < totalMemoryMatrixNumber; i++)
		{
		    t=MyReader.check("unitData/memoryMatrixes.txt", 
				     fileMemoryMatrixes, line, t);
		    if(! t.hasMoreTokens())
			lackingData("unitData/memoryMatrixes.txt");
		    matrixNumber    = Integer.parseInt(t.nextToken());
		    sensitive=true;
		    if(matrixNumber < 0){
			sensitive=false;
			matrixNumber *= -1;
		    }
		    if(matrixNumber != i){
			System.out.println("Matrix numbers in unitData/"+
					   "memoryMatrixes must be a sequence"+
					   " of natural integer number "+
					   "starting from 0.");
			MyExit.exit(1);
		    }
		    if(! t.hasMoreTokens())
			lackingData("unitData/memoryMatrixes.txt");
		    rows            = Integer.parseInt(t.nextToken());
		    if(rows <= 0){
			System.out.println("Row numbers in unitData/"+
					   "memoryMatrixes cannot be <=0.");
			MyExit.exit(1);
		    }
		    if(! t.hasMoreTokens())
			lackingData("unitData/memoryMatrixes.txt");
		    cols            = Integer.parseInt(t.nextToken());
		    if(cols <= 0){
			System.out.println("Column numbers in unitData/"+
					   "memoryMatrixes cannot be <=0.");
			MyExit.exit(1);
		    }

		    aMatrix = new MemoryMatrix(
					       getZone(),i,sensitive,
					       totalLayerNumber,rows,cols,
					       printMatrixes);
		    matrixArray.atOffset$put(i,aMatrix);
		}

	    fileMemoryMatrixes.close();
	}
	catch (FileNotFoundException fnfe)
	    {
		System.out.println(fnfe);
		MyExit.exit(1);
	    } 
	catch (IOException e)
	    {
		System.out.println("IOException:" + e.toString());
		MyExit.exit(1);
	    }

	// setting matrix designated by orMemoryMatrix in Or.java
	if(orCriterion==5){
	    if(orMemoryMatrix>(totalMemoryMatrixNumber-1)){
		System.out.println("orMemoryMatrix number cannot be greater"
				   + " than totalMemoryMatrixNumber-1; "
				   + "look at the model probe");
		MyExit.exit(1);
	    }
	    Or.setMemoryMatrix((MemoryMatrix)
			       matrixArray.atOffset(orMemoryMatrix));
	}

	} // end if(totalMemoryMatrixNumber > 0)

	return this;
    }

    /**
     * Here is where the model schedule is built, the data structures
     * that define the simulation of time in the model. The core is an
     * actionGroup that has a list of actions. That's then put in a
     * Schedule.  */
    public Object buildActions ()
    {
	int i, ii;
	super.buildActions();

	modelActions1 = new ActionGroupImpl (getZone ());
	modelActions2 = new ActionGroupImpl (getZone ());
	modelActions2generator = new ActionGroupImpl (getZone ());
	modelActions2distiller = new ActionGroupImpl (getZone ());
	modelActions2b = new ActionGroupImpl (getZone ());
	modelActions3 = new ActionGroupImpl (getZone ());

	/* zeroing daily accounting variables */
	modelActions1.createActionForEach$message(unitList,
			         SwarmUtils.getSelector("Unit","unitStep0"));
	/* clearing sequential batch production, if any */
	modelActions2.createActionForEach$message(sequentialBatchAssemblerList,
	  SwarmUtils.getSelector("SequentialBatchAssembler","clearToProduce"));
	/* producing and using inventories */
	modelActions2.createActionForEach$message(unitList,
			         SwarmUtils.getSelector("Unit","unitStep1"));
	/* a new order; this step is placed here, after the production step,
	 * to align the diffusion of the order forms (orders under execution 
	 * - next step - or new ones - this step -) */
	if(! useOrderDistiller)
	    modelActions2generator.createActionTo$message(orderGenerator,
			  SwarmUtils.getSelector(orderGenerator,
				 "createRandomOrderWithNSteps"));
	if(useOrderDistiller)
	    modelActions2distiller.createActionTo$message(orderDistiller,
			  SwarmUtils.getSelector(orderDistiller,
			  "distill"));
	/* clearing, if possible, the held order list in assigningTool
	   ('or' sequences waiting for a computaional event to apply 
	   orCriterion 5) */
	modelActions2b.
	    createActionTo$message(assigningTool,
		      SwarmUtils.getSelector(
			      assigningTool,"assignHeldOrders"));
	/* diffusing products (order forms) and making accounting per tick */
	modelActions2b.createActionForEach$message(unitList,
			  SwarmUtils.getSelector("Unit","unitStep2"));
	/* freeing the orders filled with the procurements:
	 * these are orders generated or progagated both immediately above 
	 * and in previous generation and propagation steps */
	modelActions2b.createActionForEach$message(procurementAssemblerList,
		          SwarmUtils.getSelector("ProcurementAssembler",
				     "checkingProcurementsAndFreeingOrders"));
	/* freeing the orders needing computations:
	 * these are orders generated or progagated both immediately above 
	 * and in previous generation and propagation steps */
	modelActions2b.createActionForEach$message(computationalAssemblerList,
		          SwarmUtils.getSelector("ComputationalAssembler",
				     "checkingComputationsAndFreeingOrders"));
	/* checking the orders processed in stand alone batches */
	modelActions2b.createActionForEach$message(
		       standAloneBatchAssemblerList,
		          SwarmUtils.getSelector("StandAloneBatchAssembler",
				   "checkingStandAloneBatchAndFreeingOrders"));
	/* checking the orders processed in sequential batches */
	modelActions2b.createActionForEach$message(
                       sequentialBatchAssemblerList,
		          SwarmUtils.getSelector("SequentialBatchAssembler",
				   "checkingSequentialBatchAndFreeingOrders"));
	/* propagating newses */
	modelActions2b.createActionForEach$message(unitList,
                          SwarmUtils.getSelector("Unit","unitStep3"));
	/* making daily accounting */
	modelActions3.createActionForEach$message(unitList,
			  SwarmUtils.getSelector("Unit","unitStep4"));

	// Then we create a schedule that executes the
	// modelActions. modelActions is an ActionGroup, by itself it
	// has no notion of time. In order to have it executed in
	// time, we create a Schedule that says to use the
	// modelActions ActionGroup at particular times
	modelSchedule = new ScheduleImpl (getZone (), ticksInATimeUnit);

	modelSchedule.at$createAction (0, modelActions1);

	for (i=0;i<ticksInATimeUnit;i++)
	    {
		modelSchedule.at$createAction (i, modelActions2);
		if(! useOrderDistiller)modelSchedule.
				 at$createAction (i, modelActions2generator);
		if(  useOrderDistiller && i==0)
		    for(ii=0;ii<distillerMultiplicity;ii++)
			modelSchedule.
			     at$createAction (i, modelActions2distiller);
		modelSchedule.at$createAction (i, modelActions2b);
	    }
	modelSchedule.at$createAction (ticksInATimeUnit-1, modelActions3);
	
	return this;
    }

    /**
     * Now set up the model's activation. swarmContext indicates where
     * we're being started in - typically, this model is run as a
     * subswarm of an observer swarm. */
    public Activity activateIn (Swarm swarmContext) {
	// First, activate ourselves via the superclass
	// activateIn: method.  Just pass along the context: the
	// activity library does the right thing.
	super.activateIn (swarmContext);

	// Now activate our own schedule.
	modelSchedule.activateIn (this);

	// Finally, return our activity.
	return getActivity ();
  }
    
    /** the method returns the list of the units */
    public Object getUnitList(){
	return unitList;
    }

    /** the method returns the list of the warehouses */
    public Object getWarehouseList (){
	return warehouseList;
    }

    /** the method returns the list of the orders */
    public Object getOrderList(){
        return orderList;
    }

    /** this get method returns true if warehouses are used, false otherwise */
    public boolean getUseWarehouses() {
	return useWarehouses;
    }

    /** record total production time and total recipe length */
    public void setProductionTimeAndRecipeLength(float pt, float rl) {
	totalProductionTime += pt;
	totalRecipeLength   += rl;
	if(StartESFrame.verbose)
	    System.out.println("totalProductionTime " + totalProductionTime +
			       " totalRecipeLength " + totalRecipeLength +
			       " ratio " + ((float)
					    totalProductionTime)/totalRecipeLength);
	return;
    }

    /** report the ratio totalProductionTime/totalRecipeLength */
    public float getTimeLengthRatio() {
	if(totalRecipeLength != 0)
	    return totalProductionTime/totalRecipeLength;
	else return (float) 1; // 1 is a good choice, being 0 totalRecipeLength
	                       // in the first simulation days and being 1
	                       // the actual ratio of the first
	                       // concluded orders
    }

    /** the financial rate */
    public float getInventoryFinancialRate(){
        return inventoryFinancialRate;
    }

    /** the evaluation criterion for inventories */
    public float getInventoryEvaluationCriterion(){
        return inventoryEvaluationCriterion;
    }

   /** the revenue for each recipe step */
    public float getRevenuePerEachRecipeStep(){
        return revenuePerEachRecipeStep;
    }

   /** the revenue per cost unit */
    public float getRevenuePerCostUnit(){
        return revenuePerCostUnit;
    }

    /** the number of ticks in a day */
    public int getTicksInATimeUnit(){
	return ticksInATimeUnit;
    }

    /** the benefit of the enterprise */
    public float getBenefit() {
	float benefit;
        float totalRevenues=0;
        float totalCosts=0;
        float totalSemimanufacturedRevenues=0;
        float dailyStoredComponentRevenues=0;
        float totalInventoryFinCosts=0;
        Unit aUnit;
        Order anOrder;
	int i;

	if(unitList.getCount()>0){
	    unitListIndex.setOffset(0);
	    for(i=0; i<unitList.getCount(); i++){
		aUnit=(Unit)unitListIndex.get();
		totalRevenues += aUnit.getTotalSaleRevenues();
		totalCosts += aUnit.getTotalCosts();
		if(useWarehouses){
		    dailyStoredComponentRevenues += 
			aUnit.getDailyStoredComponentValue();
		    totalInventoryFinCosts += 
			aUnit.getTotalInventoryFinancialCosts();
		}
		unitListIndex.next();
	    }
	}

	if(orderList.getCount()>0){
	    orderListIndex.setOffset(0);
	    for(i=0; i<orderList.getCount(); i++){
		anOrder = (Order)orderListIndex.get();
		totalSemimanufacturedRevenues += anOrder.getOrderEvaluation();
		orderListIndex.next();
	    }
	}

        benefit = totalRevenues + dailyStoredComponentRevenues +
	    totalSemimanufacturedRevenues - totalCosts - 
	    totalInventoryFinCosts;

        return benefit;
    }

    /** return the address of a memory matrix */
    public MemoryMatrix getMemoryMatrix(int matrixNumber){
	if(matrixNumber<0 || matrixNumber>=totalMemoryMatrixNumber)
	    {
		System.out.println("Error getting a matrix address.");
		MyExit.exit(1);
	    }

	return (MemoryMatrix) matrixArray.atOffset(matrixNumber);
    }

    /** return the number of memory matrixes */
    public int getTotalMemoryMatrixNumber(){
	return totalMemoryMatrixNumber;
	    }

    /** option about the matrix print capabulity */
    public void setMemoryMatrixPrintCapability(boolean pm){
	printMatrixes=pm;
    }

    /** conclude execution if data are lacking */
    public void lackingData(String s){
	System.out.println("Lacking data in " + s);
	MyExit.exit(1);
    }

    /** returning, as float, the number of held oders in assigningTool */
    public float getHeldOrderNumber(){
	return (float) assigningTool.getHeldOrderNumber();
    }

}


